/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.DataParsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkApplication;
import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkException;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Chromosome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataGroup;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataSet;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Genome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Probe;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.ProbeList;
import uk.ac.bbsrc.babraham.ChipMonk.GenomeParsers.EMBLParser;
import uk.ac.bbsrc.babraham.ChipMonk.GenomeParsers.GenomeParserListener;
import uk.ac.bbsrc.babraham.ChipMonk.Network.GenomeDownloader;
import uk.ac.bbsrc.babraham.ChipMonk.Network.GenomeDownloaderListener;
import uk.ac.bbsrc.babraham.ChipMonk.Network.GenomeDownloaderProgressDialog;
import uk.ac.bbsrc.babraham.ChipMonk.Normalisation.Normalisation;
import uk.ac.bbsrc.babraham.ChipMonk.Preferences.ChipMonkPreferences;

public class ChipMonkParser implements Runnable, GenomeParserListener, GenomeDownloaderListener {

	public static final int MAX_DATA_VERSION = 1;
	private ChipMonkApplication application;
	private BufferedReader br;
	private Vector listeners = new Vector();
	private DataSet [] dataSets;
	private boolean genomeLoaded = false;
	private boolean exceptionReceived = false;
	private boolean pauseWhilstLoadingGenome = false;
	private boolean requiresNormalisation = true;	
	
	public ChipMonkParser (ChipMonkApplication application) {
		this.application = application;
	}
	
	public ChipMonkApplication application () {
		return application;
	}
	
	public void parseFile (File file) {
			
		try {
			br = new BufferedReader(new FileReader(file));
		} 
		catch (FileNotFoundException ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
			}
			return;
		}
		
		Thread t = new Thread(this);
		t.start();
	}
	
	public void addDataParserListener (DataParserListener l) {
		listeners.add(l);
	}
	
	public void run() {

		genomeLoaded = false;
		exceptionReceived = false;
		
		// Loading this data is done in a modular mannar.  Every line
		// we read should be a key followed by one or more values (tab delimited).
		// Whatever it is we pass it on to a specialised subroutine to deal with.

		try {
			// The first line is the genome we're dealing with
			String line;
			String [] sections;
			while ((line = br.readLine()) != null) {
				sections = line.split("\\t");
				
				// Now we look where to send this...
				if (sections[0].equals("Chipmonk Data Version")) {
					parseDataVersion(sections);
				}				
				else if (sections[0].equals("Features")) {
					parseFeatures(sections);
				}
				else if (sections[0].equals("Normalisations")) {
					parseNormalisations(sections);
				}
				else if (sections[0].equals("Samples")) {
					parseSamples(sections);
				}
				else if (sections[0].equals("Data Groups")) {
					parseGroups(sections);
				}
				else if (sections[0].equals("Probes")) {
					parseProbes(sections);
				}
				else if (sections[0].equals("Lists")) {
					parseLists(sections);
				}
				else if (sections[0].equals("Genome")) {
					parseGenome(sections);
				}
				else if (sections[0].equals("Visible Stores")) {
					parseVisibleStores(sections);
				}
				else {
					throw new ChipMonkException("Didn't recogise section '"+sections[0]+"' in chipmonk file");
				}
			}
			
			// We're finished with the file
			br.close();

		}
		catch (Exception ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
			}
			return;
		}
				
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((DataParserListener)e.nextElement()).dataProcessingFinished(requiresNormalisation);
		}
		application.resetChangesWereMade();
	
	}

	private void parseDataVersion (String [] sections) throws ChipMonkException {
		if (sections.length != 2) {
			throw new ChipMonkException("Data Version line didn't contain 2 sections");
		}
		
		int v = Integer.parseInt(sections[1]);
		
		if (v > MAX_DATA_VERSION) {
			throw new ChipMonkException("This data file needs a newer verison of ChIPMonk to read it.");
		}
	}
	
	private void parseGenome (String [] sections) throws ChipMonkException {
		if (sections.length != 3) {
			throw new ChipMonkException("Genome line didn't contain 3 sections");
		}
		if (! sections[0].equals("Genome")){
			throw new ChipMonkException("First line of file was not the genome description");
		}
		
		File f;
		try {
			f = new File(ChipMonkPreferences.getInstance().getGenomeBase().getAbsoluteFile()+"/"+sections[1]+"/"+sections[2]);
		} 
		catch (FileNotFoundException e) {
			throw new ChipMonkException("Couldn't find the folder which should contain the genome files.  Please check your file preferences.");
		}
		
		if (! f.exists()) {
			// The user doesn't have this genome - yet...
			GenomeDownloader d = new GenomeDownloader();
			d.addGenomeDownloaderListener(this);
			GenomeDownloaderProgressDialog pd = new GenomeDownloaderProgressDialog(application);
			d.addGenomeDownloaderListener(pd);
			// Bit of a hack here, since we don't know the size of the genome
			// being downloaded we have to put an approximate value in so the
			// progress bar does something sensible.
			d.downloadGenome(sections[1],sections[2],25000000);
			pd.requestFocus();
			pauseWhilstLoadingGenome = true;
			while (pauseWhilstLoadingGenome) {
				if (exceptionReceived) throw new ChipMonkException("Unable to download specified genome");
				try {
					Thread.sleep(1000);
				}
				catch (InterruptedException ie) {}
			}
			
		}
		EMBLParser parser = new EMBLParser();
		parser.addGenomeParserListener(this);
		parser.parseGenome(f);

		while (! genomeLoaded) {
			if (exceptionReceived) return;
			try {
				Thread.sleep(1000);
			}
			catch (InterruptedException ie) {}
		}

	}
	
	private void parseFeatures (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Features line didn't contain 2 sections");
		}
		if (! sections[0].equals("Features")){
			throw new ChipMonkException("Couldn't find expected features line");
		}
		int n = Integer.parseInt(sections[1]);
		
		Vector defaultFeaturesVector = new Vector();
		
		for (int i=0;i<n;i++) {
			String type = br.readLine();
			if (ChipMonkPreferences.getInstance().loadAnnotation(type)) {
				defaultFeaturesVector.add(type);
			}
		}

			
		application.setDrawnFeatureTypes((String [])defaultFeaturesVector.toArray(new String[0]));
	}
	
	private void parseNormalisations (String [] sections) throws InstantiationException, IllegalAccessException, ClassNotFoundException, IOException, ChipMonkException  {

		if (sections.length != 2) {
			throw new ChipMonkException("Normalisation line didn't contain 2 sections");
		}
		if (! sections[0].equals("Normalisations")){
			throw new ChipMonkException("Couldn't find expected normalisation line");
		}
		int n = Integer.parseInt(sections[1]);
		
		Normalisation [] norms = new Normalisation[n];
		
		for (int i=0;i<n;i++) {
			norms[i] = (Normalisation)Class.forName(br.readLine()).newInstance();
		}
		
		// This might be causing problems with two modules trying
		// to normalise the data at the same time.
		application.setNormalisationModulesWithoutApplying(norms);
	}
	
	private void parseSamples (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Samples line didn't contain 2 sections");
		}
		if (! sections[0].equals("Samples")){
			throw new ChipMonkException("Couldn't find expected samples line");
		}
		
		int n = Integer.parseInt(sections[1]);
		
		// We need to keep the Data Sets around to add data to later.
		dataSets = new DataSet[n];
		
		for (int i=0;i<n;i++) {
			dataSets[i] = new DataSet(br.readLine());
			application.dataCollection().addDataSet(dataSets[i]);
		}
	}
	
	private void parseGroups (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Data Groups line didn't contain 2 sections");
		}
		if (! sections[0].equals("Data Groups")){
			throw new ChipMonkException("Couldn't find expected data groups line");
		}
		
		int n = Integer.parseInt(sections[1]);
		
		for (int i=0;i<n;i++) {
			String [] group = br.readLine().split("\\t");
			DataSet [] groupMembers = new DataSet[group.length-1];
			for (int j=1;j<group.length;j++) {
				groupMembers[j-1] = application.dataCollection().getDataSet(group[j]);
			}
			DataGroup g = new DataGroup(group[0],groupMembers);
			application.dataCollection().addDataGroup(g);
		}
	}
	
	private void parseProbes (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Probe line didn't contain 2 sections");
		}
		if (! sections[0].equals("Probes")){
			throw new ChipMonkException("Couldn't find expected probes line");
		}
		
		int n = Integer.parseInt(sections[1]);
		
		boolean includeNorm = false;
		int expectedSectionLength = -1;
		
		for (int i=0;i<n;i++) {
			String line = br.readLine();
			if (line == null) {
				throw new ChipMonkException("Ran out of probe data at line "+i+" (expected "+n+" probes)");
			}
			sections = line.split("\\t");

			if (i==0) {
				// On the first line we need to decide if we're going to read
				// normalised data as well.

				if (sections.length == (dataSets.length*2)+3) {
					includeNorm = false;
					requiresNormalisation = true;
					expectedSectionLength = sections.length;
				}
				else if (sections.length == (dataSets.length*4)+3) {
					includeNorm = true;
					requiresNormalisation = false;
					expectedSectionLength = sections.length;
				}
				else {
					throw new ChipMonkException("Number of data sections ("+sections.length+") was wrong for either 2 ("+((dataSets.length*2)+3)+") or 4 ("+((dataSets.length*4)+3)+") values per probe");
				}
			}

			
			if (sections.length != expectedSectionLength) {
				throw new ChipMonkException ("Expected "+expectedSectionLength+" sections in data file for "+sections[0]+" but got "+sections.length);
			}
			
			if (i%1000 == 0) {
				Enumeration e = listeners.elements();
				while (e.hasMoreElements()) {
					((DataParserListener)e.nextElement()).dataUpdateProgress(i,n,"Processed data for "+i+" probes");
				}					
			}
			Chromosome c = application.genome().getChromosome(sections[1]);
			int position = Integer.parseInt(sections[2]);
			if (c == null) {
				Enumeration e = listeners.elements();
				while (e.hasMoreElements()) {
					((DataParserListener)e.nextElement()).dataWarningReceived(new ChipMonkException("Probe "+sections[0]+" could not be mapped to chromosome '"+sections[1]+"'"));
				}
			}
			else if (position > c.length()) {
				Enumeration e = listeners.elements();
				while (e.hasMoreElements()) {
					((DataParserListener)e.nextElement()).dataWarningReceived(new ChipMonkException("Probe "+sections[0]+" has a position "+position+" beyond the end of chr"+c.name()+" ("+c.length()+")"));
				}
				c = null;
				position = 0;
			}
			
			
			Probe p = new Probe(sections[0],c,position);
			application.dataCollection().probeSet().addProbe(p);

			for (int j=0;j<dataSets.length;j++) {
				if (includeNorm) {
					dataSets[j].addData(p,Double.parseDouble(sections[3+(4*j)]),Double.parseDouble(sections[4+(4*j)]),Double.parseDouble(sections[5+(4*j)]),Double.parseDouble(sections[6+(4*j)]));
				}
				else {
					dataSets[j].addData(p,Double.parseDouble(sections[3+(2*j)]),Double.parseDouble(sections[4+(2*j)]));
				}
			}
		}
	}
	
	private void parseVisibleStores (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Visible stores line didn't contain 2 sections");
		}
		int n = Integer.parseInt(sections[1]);

		for (int i=0;i<n;i++) {
			String line = br.readLine();
			if (line == null) {
				throw new ChipMonkException("Ran out of visible store data at line "+i+" (expected "+n+" stores)");
			}
			String [] storeSections = line.split("\\t");
			if (storeSections.length != 2) {
				throw new ChipMonkException("Expected 2 sections in visible store line but got "+storeSections.length);
			}
			if (storeSections[1].equals("set")) {
				application.addToDrawnDataStores(application.dataCollection().getDataSet(storeSections[0]));
			}
			else if (storeSections[1].equals("group")) {
				application.addToDrawnDataStores(application.dataCollection().getDataGroup(storeSections[0]));
			}
			else {
				throw new ChipMonkException("Didn't recognise data type '"+storeSections[1]+"' when adding visible stores from line '"+line+"'");
			}
		}
			
	}
	
	private void parseLists (String [] sections) throws ChipMonkException, IOException {
		if (sections.length != 2) {
			throw new ChipMonkException("Probe Lists line didn't contain 2 sections");
		}
		
		int n = Integer.parseInt(sections[1]);
		ProbeList [] lists = new ProbeList[n];
		for (int i=0;i<n;i++) {
			String line = br.readLine();
			if (line == null) {
				throw new ChipMonkException("Ran out of probe data at line "+i+" (expected "+n+" probes)");
			}
			String [] listSections = line.split("\\t");
			lists[i] = new ProbeList(application.dataCollection().probeSet(),listSections[0],listSections[1]);
			application.dataCollection().probeSet().addProbeList(lists[i]);
		}
		
		// Next we reach the probe list data.  These comes as a long list of values
		// the first of which is the probe name, then either a numerical value if
		// the probe is contained in that list, or a blank if it isn't.
		String line = br.readLine();
		if (line == null) {
			throw new ChipMonkException("Couldn't find probe line for list data");				
		}
		sections = line.split("\\t");
		if (sections.length != 2) {
			throw new ChipMonkException("Probe line didn't contain 2 sections");
		}
		if (! sections[0].equals("Probes")){
			throw new ChipMonkException("Couldn't find expected probe lists probe line");
		}
		
		n = Integer.parseInt(sections[1]);
		
		for (int i=0;i<n;i++) {
			sections = br.readLine().split("\\t",-1);
			if (sections.length != lists.length+1) {
				throw new ChipMonkException ("Expected "+(lists.length+1)+" sections in probe list section of data file for "+sections[0]+" but got "+sections.length);
			}
			
			if (i%1000 == 0) {
				Enumeration e = listeners.elements();
				while (e.hasMoreElements()) {
					((DataParserListener)e.nextElement()).dataUpdateProgress(i,n,"Processed list data for "+i+" probes");
				}					
			}

			Probe p = application.dataCollection().probeSet().getProbeWithName(sections[0]);
				

			for (int j=0;j<lists.length;j++) {
				if (sections[j+1].length()>0) {
					lists[j].addProbe(p,new Double(Double.parseDouble(sections[j+1])));
				}
			}
		}
	}
		
	public void genomeUpdateProgress(int currentPos, int totalPos, String message) {
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((DataParserListener)e.nextElement()).dataUpdateProgress(currentPos, totalPos, message);
		}
	}

	public void genomeExceptionReceived(Exception ex) {
		
		exceptionReceived = true;
		
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
		}
		
	}

	public void genomeProcessingFinished(Genome g) {
		application.genomeProcessingFinished(g);
		genomeLoaded = true;
	}

	public void downloadUpdateProgress(int currentPos, int totalPos, String message) {		
	}
	public void downloadExceptionReceived(Exception ex) {
		exceptionReceived = true;
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
		}
	}

	public void downloadProcessingFinished() {
		pauseWhilstLoadingGenome = false;
	}

}
