/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.DataWriters;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Vector;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkApplication;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataCollection;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataGroup;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataSet;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataStore;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Genome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Probe;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.ProbeList;
import uk.ac.bbsrc.babraham.ChipMonk.Normalisation.Normalisation;

public class ChipMonkDataWriter implements Runnable {
	
	// THIS VALUE IS IMPORTANT!!!
	public static final int DATA_VERSION = 1;
	
	// If you make ANY changes to the format written by this class
	// you MUST increment this value to stop older parsers from
	// trying to parse it.  Once you have updated the parser to 
	// read the new format you can then update the corresponding
	// value in the parser so that it will work.
	
	private Vector listeners = new Vector();
	private DataCollection data;
	private Genome genome;
	private File file;
	private Normalisation [] normalisations;
	private String [] defaultFeatureTracks;
	private DataStore [] visibleStores;
	
	public ChipMonkDataWriter () {
	}
	
	public void addDataWriterListener (DataWriterListener l) {
		listeners.add(l);
	}
	
	public void writeData (ChipMonkApplication application, File file) {
		data = application.dataCollection();
		genome = application.genome();
		this.file = file;
		normalisations = application.getNormalisationModules();
		defaultFeatureTracks = application.drawnFeatureTypes();
		visibleStores = application.drawnDataStores();
		
		Thread t = new Thread(this);
		t.start();
	}

	public void run() {
		try {
			PrintStream p = new PrintStream(new FileOutputStream(file));
			
			// The first line of the file will be the version of the data
			// format we're using.  This will help us out should we need 
			// to update the format in the future.
			p.println("Chipmonk Data Version\t"+DATA_VERSION);
			
			// The next thing we need to do is to output the details of the genome
			// we're using
			p.println("Genome\t"+genome.species()+"\t"+genome.assembly());
			
			// Now we print out the list of default feature tracks to show
			p.println("Features\t"+defaultFeatureTracks.length);
			for (int i=0;i<defaultFeatureTracks.length;i++) {
				p.println(defaultFeatureTracks[i]);
			}
			
			// Now we print out the list of normalisations we're using
			p.println("Normalisations\t"+normalisations.length);
			
			for (int i=0;i<normalisations.length;i++) {
				p.println(normalisations[i].getClass().getName());
			}
			
			// Now we print out the list of datasets
			DataSet [] dataSets = data.getAllDataSets();
			p.println("Samples\t"+dataSets.length);
			for (int i=0;i<dataSets.length;i++) {
				p.println(dataSets[i].name());
			}
			
			//Now we print out the list of data groups
			DataGroup [] dataGroups = data.getAllDataGroups();
			p.println("Data Groups\t"+dataGroups.length);
			for (int i=0;i<dataGroups.length;i++) {
				DataSet [] groupSets = dataGroups[i].dataSets();
				StringBuffer b = new StringBuffer();
				b.append(dataGroups[i].name());
				for (int j=0;j<groupSets.length;j++) {
					b.append("\t");
					b.append(groupSets[j].name());
				}
				
				p.println(b);
			}
			

			//Put out the number of probes
			Probe [] probes = data.probeSet().getAllProbes();
			p.println("Probes\t"+probes.length);

			// Next we print out the data

			for (int i=0;i<probes.length;i++) {
				
				if (i%1000 == 0) {
					Enumeration e = listeners.elements();
					while (e.hasMoreElements()) {
						((DataWriterListener)e.nextElement()).dataWriterProgressUpdate("Written data for "+i+" probes out of "+probes.length,i,probes.length);
					}
				}
				
				StringBuffer b = new StringBuffer();
				b.append(probes[i].name());
				b.append("\t");
				if (probes[i].chromosome() != null)
					b.append(probes[i].chromosome().name());
				b.append("\t");
				b.append(probes[i].position());
				
				for (int j=0;j<dataSets.length;j++) {
					b.append("\t");
					b.append(dataSets[j].getRawSignalValueForProbe(probes[i]));
					b.append("\t");					
					b.append(dataSets[j].getRawControlValueForProbe(probes[i]));
					b.append("\t");
					b.append(dataSets[j].getSignalValueForProbe(probes[i]));
					b.append("\t");					
					b.append(dataSets[j].getControlValueForProbe(probes[i]));
					
				}
				p.println(b.toString());
			}
			
			// Now we can put out the list of visible stores
			p.println("Visible Stores\t"+visibleStores.length);
			for (int i=0;i<visibleStores.length;i++) {
				String type;
				if (visibleStores[i] instanceof DataSet) {
					type = "set";
				}
				else {
					type = "group";
				}
				p.println(visibleStores[i].name()+"\t"+type);
			}
			
			// Now we print out the list of probe lists
			ProbeList [] lists = data.probeSet().getAllProbeLists();
			p.println("Lists\t"+lists.length);
			for (int i=0;i<lists.length;i++) {
				p.println(lists[i].getListName()+"\t"+lists[i].getValueName());
			}
			
			//Put out the number of probes
			p.println("Probes\t"+probes.length);
			// Now we print out the data for the probe lists
			for (int i=0;i<probes.length;i++) {
				
				if (i%1000 == 0) {
					Enumeration e = listeners.elements();
					while (e.hasMoreElements()) {
						((DataWriterListener)e.nextElement()).dataWriterProgressUpdate("Written lists for "+i+" probes out of "+probes.length,i,probes.length);
					}
				}
				
				StringBuffer b = new StringBuffer();
				b.append(probes[i].name());
				
				for (int j=0;j<lists.length;j++) {
					b.append("\t");
					if (lists[j].containsProbe(probes[i])) {
						b.append(lists[j].getValueForProbe(probes[i]));
					}
				}
				p.println(b.toString());
			}
			
			
			p.close();
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataWriterListener)e.nextElement()).dataWritingComplete();
			}
		}
		
		catch (Exception ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataWriterListener)e.nextElement()).dataWriterExceptionReceived(ex);
			}
		}
		
	}

	
}
