/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.Preferences;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.swing.JOptionPane;

public class ChipMonkPreferences {

	
	private static ChipMonkPreferences p = new ChipMonkPreferences();
	
	private Hashtable defaultAnnotations = new Hashtable();
	private File genomeBase = null;
	private File dataLocation = new File(System.getProperty("user.home"));
	private File saveLocation = new File(System.getProperty("user.home"));
	// Because this value is going to be used to create a hashtable
	// it's better if it's a prime as this reduces collisions.
	private int approxProbeSetSize = 500009;
	private File preferencesFile = null;
	private String genomeDownloadLocation = "http://www.bioinformatics.bbsrc.ac.uk/chipmonk/genomes/";
	private boolean useProxy = false;
	private String proxyHost = "";
	private int proxyPort = 0;
	
	private ChipMonkPreferences () {

		// Check for the default genomes folder.  This should always be present, but
		// you can't be too careful!
		try {
			genomeBase = new File(ClassLoader.getSystemResource("Genomes").getFile().replaceAll("%20"," "));
		}
		catch (NullPointerException npe) {
			JOptionPane.showMessageDialog(null, "Couldn't find the 'Genomes' directory in your chipmonk installation", "Warning", JOptionPane.WARNING_MESSAGE);
		}

		
		//		System.out.println("Looking for preferences at: "+System.getProperty("user.home")+"/chipmonk_prefs.txt");
		preferencesFile= new File(System.getProperty("user.home")+"/chipmonk_prefs.txt");
		
		if (preferencesFile!=null && preferencesFile.exists()) {
//			System.out.println("Loading preferences from file...");
			loadPreferences();
		}
		else {
			defaultAnnotations.put("cds","");
			defaultAnnotations.put("mrna","");
			defaultAnnotations.put("gene","");
			try {
				savePreferences();
			} 
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		updateProxyInfo();
	}
	
	private void loadPreferences () {

		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(preferencesFile));
		

			String line;
			String [] sections;
			while ((line = br.readLine())!=null) {
				sections = line.split("\\t");
				
				if (sections[0].equals("GenomeBase")) {
					genomeBase = new File (sections[1]);
				}
				else if (sections[0].equals("DataLocation")) {
					dataLocation = new File(sections[1]);
				}
				else if (sections[0].equals("SaveLocation")) {
					saveLocation = new File(sections[1]);
				}				
				else if (sections[0].equals("ProbesetSize")) {
					approxProbeSetSize = Integer.parseInt(sections[1]);
				}
				else if (sections[0].equals("GenomeDownloadLocation")) {
					genomeDownloadLocation = sections[1];
				}
				else if (sections[0].equals("Proxy")) {
					proxyHost = sections[1];
					if (sections.length>2 && sections[2].length()>0) {
						proxyPort = Integer.parseInt(sections[2]);
					}
					if (proxyHost.length()>0) {
						useProxy = true;
					}
				}
				else if (sections[0].equals("LoadedFeatures")) {
					for (int i=1;i<sections.length;i++) {
						defaultAnnotations.put(sections[i].toLowerCase(),"");
					}
				}
				
			}
			br.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	
	}
	
	public void savePreferences() throws IOException {
		PrintWriter p = new PrintWriter(new FileWriter(preferencesFile));
		
		p.println("# ChIPMonnk Preferences file.  Do not edit by hand.");
		
		// First write out the GenomeBase
		p.println("GenomeBase\t"+genomeBase.getAbsolutePath());

		// Then the dataLocation
		p.println("DataLocation\t"+dataLocation.getAbsolutePath());
		
		// Then the saveLocation
		p.println("SaveLocation\t"+saveLocation.getAbsolutePath());
		
		// Then the probeset size
		p.println("ProbesetSize\t"+approxProbeSetSize);
				
		// Then the proxy information
		p.println("Proxy\t"+proxyHost+"\t"+proxyPort);
		
		// The genome download URL
		p.println("GenomeDownloadLocation\t"+genomeDownloadLocation);
		
		// Finally the list of features to load
		StringBuffer b = new StringBuffer("LoadedFeatures");
		Enumeration e = defaultAnnotations.keys();
		while (e.hasMoreElements()) {
			b.append("\t");
			b.append(e.nextElement());
		}
		p.println(b);
		
		p.close();
		
	}
	
	public static ChipMonkPreferences getInstance () {		
		return p;
	}
	
	public boolean loadAnnotation (String type) {
		if (defaultAnnotations.size() > 0 && (! defaultAnnotations.containsKey(type.toLowerCase()))) {
			return false;
		}
		else {
			return true;
		}
	}
	
	public String [] getLoadedFeatures () {
		String [] f = new String [defaultAnnotations.size()];
		
		Enumeration e = defaultAnnotations.keys();
		int i=0;
		while (e.hasMoreElements()) {
			f[i] = (String)e.nextElement();
			i++;
		}
		
		return f;
	}
	
	public boolean useProxy () {
		return useProxy;
	}
	
	public String proxyHost () {
		return proxyHost;
	}
	
	public int proxyPort () {
		return proxyPort;
	}
	
	public void setProxy (String host, int port) {
		proxyHost = host;
		proxyPort = port;
		updateProxyInfo();
	}
	
	public void setGenomeDownloadLocation (String url) {
		genomeDownloadLocation = url;
	}
	
	public File getGenomeBase () throws FileNotFoundException {
		if (genomeBase == null || !genomeBase.exists()) {
			throw new FileNotFoundException("Couldn't find genomes folder location.  Please check your preferences.");
		}
		return genomeBase;
	}
	
	public String getGenomeDownloadLocation () {
		return genomeDownloadLocation;
	}
	
	public void setGenomeBase (File f) {
		genomeBase = f;
	}
	
	public File getDataLocation () {
		return dataLocation;
	}
	
	public void setDataLocation (File f) {
		dataLocation = f;
	}
	
	public File getSaveLocation () {
		return saveLocation;
	}
	
	public void setSaveLocation (File f) {
		saveLocation = f;
	}
	
	public int getApproxProbeSetSize () {
		return approxProbeSetSize;
	}
	
	public void setDefaultAnnotations (String [] s) {
		defaultAnnotations = new Hashtable();
		for (int i=0;i<s.length;i++) {
			defaultAnnotations.put(s[i],"");
		}
	}
	
	public void setApproxProbeSetSize (int i) {
		approxProbeSetSize = i;
	}
		
	private void updateProxyInfo () {
		if (useProxy) {
			System.getProperties().put("proxySet","true");
			System.getProperties().put("proxyHost",proxyHost);
			System.getProperties().put("proxyPort",""+proxyPort);
		}
		else {
			System.getProperties().put("proxySet","false");
		}
	}
}
